"""Deep Feature Kernel Density Estimation model."""

# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

from .lightning_model import Dfkde, DfkdeLightning

__all__ = ["Dfkde", "DfkdeLightning"]
