from setuptools import setup, find_packages

setup(
    name='fhailib',
    version='1.0.1',
    description='FH,FJ Vision AI SYSTEM LIB',
    author='LEE_DH',
    author_email='ldh9616@nate.com',
    url='https://github.com/LEE-pyt/AI_Module.git',
    install_requires=['tqdm', 'pandas', 'scikit-learn','torch==2.0.1','torchvision==0.15.2','onnx','onnxruntime','pycocotools','anomalib','GitPython','PyYAML','albumentations>=1.1.0','av>=10.0.0','comet-ml>=3.31.7','coverage[toml]','defusedxml==0.7.1','einops>=0.3.2','freia>=0.2','furo==2022.9.29','gitpython','gradio>=2.9.4','imgaug==0.4.0','ipykernel','ipywidgets','jsonargparse[signatures]>=4.3','kornia>=0.6.6,<0.6.10','matplotlib>=3.4.3','myst-parser','nbsphinx>=0.8.9','networkx~=2.5','nncf>=2.1.0','notebook','omegaconf>=2.1.1','onnx==1.14.0','onnxruntime==1.15.1','opencv-python>=4.5.3.56','openvino-dev>=2022.3.0','pandas>=1.1.0','pandoc','pre-commit','pycocotools','pytest','pytest-cov','pytest-order','pytest-sugar','pytest-xdist','pytorch-lightning>=1.7.0,<1.10.0','requests>=2.26.0','scipy','setuptools>=41.0.0','sphinx>=4.1.2','sphinx-autoapi','sphinxemoji==0.1.8','tensorboard','timm>=0.5.4,<=0.6.13','torch==2.0.1','torchmetrics==0.10.3','torchvision==0.15.2','tox','wandb==0.12.17'],
    packages=find_packages(exclude=[]),
    keywords=['FH_AI', 'FH_AI_TOOL', 'FH_AI_LIB', 'python tutorial', 'pypi'],
    python_requires='>=3.8',
    package_data={},
    zip_safe=False,
    py_modules=['AI_Module.anomal_util.anomal_Tool', 'AI_Module.anomal_util.anomalib.config.config', 'AI_Module.anomal_util.anomalib.data.avenue', 'AI_Module.anomal_util.anomalib.data.btech', 'AI_Module.anomal_util.anomalib.data.folder', 'AI_Module.anomal_util.anomalib.data.folder_3d', 'AI_Module.anomal_util.anomalib.data.inference', 'AI_Module.anomal_util.anomalib.data.kolektor', 'AI_Module.anomal_util.anomalib.data.mvtec', 'AI_Module.anomal_util.anomalib.data.mvtec_3d', 'AI_Module.anomal_util.anomalib.data.shanghaitech', 'AI_Module.anomal_util.anomalib.data.synthetic', 'AI_Module.anomal_util.anomalib.data.task_type', 'AI_Module.anomal_util.anomalib.data.ucsd_ped', 'AI_Module.anomal_util.anomalib.data.visa', 'AI_Module.anomal_util.anomalib.data.base.datamodule', 'AI_Module.anomal_util.anomalib.data.base.dataset', 'AI_Module.anomal_util.anomalib.data.base.depth', 'AI_Module.anomal_util.anomalib.data.base.video', 'AI_Module.anomal_util.anomalib.data.utils.augmenter', 'AI_Module.anomal_util.anomalib.data.utils.boxes', 'AI_Module.anomal_util.anomalib.data.utils.download', 'AI_Module.anomal_util.anomalib.data.utils.image', 'AI_Module.anomal_util.anomalib.data.utils.label', 'AI_Module.anomal_util.anomalib.data.utils.path', 'AI_Module.anomal_util.anomalib.data.utils.split', 'AI_Module.anomal_util.anomalib.data.utils.transform', 'AI_Module.anomal_util.anomalib.data.utils.video', 'AI_Module.anomal_util.anomalib.data.utils.generators.perlin', 'AI_Module.anomal_util.anomalib.deploy.export', 'AI_Module.anomal_util.anomalib.deploy.inferencers.base_inferencer', 'AI_Module.anomal_util.anomalib.deploy.inferencers.openvino_inferencer', 'AI_Module.anomal_util.anomalib.deploy.inferencers.torch_inferencer', 'AI_Module.anomal_util.anomalib.models.ai_vad.density', 'AI_Module.anomal_util.anomalib.models.ai_vad.features', 'AI_Module.anomal_util.anomalib.models.ai_vad.flow', 'AI_Module.anomal_util.anomalib.models.ai_vad.lightning_model', 'AI_Module.anomal_util.anomalib.models.ai_vad.regions', 'AI_Module.anomal_util.anomalib.models.ai_vad.torch_model', 'AI_Module.anomal_util.anomalib.models.ai_vad.clip.clip', 'AI_Module.anomal_util.anomalib.models.ai_vad.clip.model', 'AI_Module.anomal_util.anomalib.models.cfa.anomaly_map', 'AI_Module.anomal_util.anomalib.models.cfa.lightning_model', 'AI_Module.anomal_util.anomalib.models.cfa.loss', 'AI_Module.anomal_util.anomalib.models.cfa.torch_model', 'AI_Module.anomal_util.anomalib.models.cflow.anomaly_map', 'AI_Module.anomal_util.anomalib.models.cflow.lightning_model', 'AI_Module.anomal_util.anomalib.models.cflow.torch_model', 'AI_Module.anomal_util.anomalib.models.cflow.utils', 'AI_Module.anomal_util.anomalib.models.components.base.anomaly_module', 'AI_Module.anomal_util.anomalib.models.components.base.dynamic_module', 'AI_Module.anomal_util.anomalib.models.components.classification.kde_classifier', 'AI_Module.anomal_util.anomalib.models.components.cluster.kmeans', 'AI_Module.anomal_util.anomalib.models.components.dimensionality_reduction.pca', 'AI_Module.anomal_util.anomalib.models.components.dimensionality_reduction.random_projection', 'AI_Module.anomal_util.anomalib.models.components.feature_extractors.timm', 'AI_Module.anomal_util.anomalib.models.components.feature_extractors.torchfx', 'AI_Module.anomal_util.anomalib.models.components.feature_extractors.utils', 'AI_Module.anomal_util.anomalib.models.components.filters.blur', 'AI_Module.anomal_util.anomalib.models.components.flow.all_in_one_block', 'AI_Module.anomal_util.anomalib.models.components.layers.sspcab', 'AI_Module.anomal_util.anomalib.models.components.sampling.k_center_greedy', 'AI_Module.anomal_util.anomalib.models.components.stats.kde', 'AI_Module.anomal_util.anomalib.models.components.stats.multi_variate_gaussian', 'AI_Module.anomal_util.anomalib.models.csflow.anomaly_map','AI_Module.anomal_util.anomalib.models.csflow.lightning_model', 'AI_Module.anomal_util.anomalib.models.csflow.loss', 'AI_Module.anomal_util.anomalib.models.csflow.torch_model', 'AI_Module.anomal_util.anomalib.models.dfkde.lightning_model', 'AI_Module.anomal_util.anomalib.models.dfkde.torch_model', 'AI_Module.anomal_util.anomalib.models.dfm.lightning_model', 'AI_Module.anomal_util.anomalib.models.dfm.torch_model', 'AI_Module.anomal_util.anomalib.models.draem.lightning_model', 'AI_Module.anomal_util.anomalib.models.draem.loss', 'AI_Module.anomal_util.anomalib.models.draem.torch_model', 'AI_Module.anomal_util.anomalib.models.dsr.anomaly_generator', 'AI_Module.anomal_util.anomalib.models.dsr.lightning_model', 'AI_Module.anomal_util.anomalib.models.dsr.loss', 'AI_Module.anomal_util.anomalib.models.dsr.torch_model', 'AI_Module.anomal_util.anomalib.models.efficient_ad.lightning_model', 'AI_Module.anomal_util.anomalib.models.efficient_ad.torch_model', 'AI_Module.anomal_util.anomalib.models.fastflow.anomaly_map', 'AI_Module.anomal_util.anomalib.models.fastflow.lightning_model', 'AI_Module.anomal_util.anomalib.models.fastflow.loss', 'AI_Module.anomal_util.anomalib.models.fastflow.torch_model', 'AI_Module.anomal_util.anomalib.models.ganomaly.lightning_model', 'AI_Module.anomal_util.anomalib.models.ganomaly.loss', 'AI_Module.anomal_util.anomalib.models.ganomaly.torch_model', 'AI_Module.anomal_util.anomalib.models.padim.anomaly_map', 'AI_Module.anomal_util.anomalib.models.padim.lightning_model', 'AI_Module.anomal_util.anomalib.models.padim.torch_model', 'AI_Module.anomal_util.anomalib.models.patchcore.anomaly_map', 'AI_Module.anomal_util.anomalib.models.patchcore.lightning_model', 'AI_Module.anomal_util.anomalib.models.patchcore.torch_model', 'AI_Module.anomal_util.anomalib.models.reverse_distillation.anomaly_map', 'AI_Module.anomal_util.anomalib.models.reverse_distillation.lightning_model', 'AI_Module.anomal_util.anomalib.models.reverse_distillation.loss', 'AI_Module.anomal_util.anomalib.models.reverse_distillation.torch_model', 'AI_Module.anomal_util.anomalib.models.reverse_distillation.components.bottleneck', 'AI_Module.anomal_util.anomalib.models.reverse_distillation.components.de_resnet', 'AI_Module.anomal_util.anomalib.models.rkde.feature_extractor', 'AI_Module.anomal_util.anomalib.models.rkde.lightning_model', 'AI_Module.anomal_util.anomalib.models.rkde.region_extractor', 'AI_Module.anomal_util.anomalib.models.rkde.torch_model', 'AI_Module.anomal_util.anomalib.models.stfpm.anomaly_map', 'AI_Module.anomal_util.anomalib.models.stfpm.lightning_model', 'AI_Module.anomal_util.anomalib.models.stfpm.loss', 'AI_Module.anomal_util.anomalib.models.stfpm.torch_model', 'AI_Module.anomal_util.anomalib.models.uflow.anomaly_map', 'AI_Module.anomal_util.anomalib.models.uflow.feature_extraction', 'AI_Module.anomal_util.anomalib.models.uflow.lightning_model', 'AI_Module.anomal_util.anomalib.models.uflow.loss', 'AI_Module.anomal_util.anomalib.models.uflow.torch_model', 'AI_Module.anomal_util.anomalib.post_processing.post_process', 'AI_Module.anomal_util.anomalib.post_processing.visualizer', 'AI_Module.anomal_util.anomalib.post_processing.normalization.cdf', 'AI_Module.anomal_util.anomalib.post_processing.normalization.min_max', 'AI_Module.anomal_util.anomalib.pre_processing.pre_process', 'AI_Module.anomal_util.anomalib.pre_processing.tiler', 'AI_Module.anomal_util.anomalib.pre_processing.transforms.custom', 'AI_Module.anomal_util.anomalib.utils.callbacks.cdf_normalization', 'AI_Module.anomal_util.anomalib.utils.callbacks.export', 'AI_Module.anomal_util.anomalib.utils.callbacks.graph', 'AI_Module.anomal_util.anomalib.utils.callbacks.metrics_configuration', 'AI_Module.anomal_util.anomalib.utils.callbacks.min_max_normalization', 'AI_Module.anomal_util.anomalib.utils.callbacks.model_loader', 'AI_Module.anomal_util.anomalib.utils.callbacks.post_processing_configuration', 'AI_Module.anomal_util.anomalib.utils.callbacks.tiler_configuration', 'AI_Module.anomal_util.anomalib.utils.callbacks.timer', 'AI_Module.anomal_util.anomalib.utils.callbacks.nncf.callback', 'AI_Module.anomal_util.anomalib.utils.callbacks.nncf.utils', 'AI_Module.anomal_util.anomalib.utils.callbacks.visualizer.visualizer_base', 'AI_Module.anomal_util.anomalib.utils.callbacks.visualizer.visualizer_image', 'AI_Module.anomal_util.anomalib.utils.callbacks.visualizer.visualizer_metric', 'AI_Module.anomal_util.anomalib.utils.cli.cli', 'AI_Module.anomal_util.anomalib.utils.cv.connected_components', 'AI_Module.anomal_util.anomalib.utils.hpo.config', 'AI_Module.anomal_util.anomalib.utils.hpo.runners', 'AI_Module.anomal_util.anomalib.utils.loggers.base', 'AI_Module.anomal_util.anomalib.utils.loggers.comet', 'AI_Module.anomal_util.anomalib.utils.loggers.tensorboard', 'AI_Module.anomal_util.anomalib.utils.loggers.wandb', 'AI_Module.anomal_util.anomalib.utils.metrics.anomaly_score_distribution', 'AI_Module.anomal_util.anomalib.utils.metrics.anomaly_score_threshold', 'AI_Module.anomal_util.anomalib.utils.metrics.aupr', 'AI_Module.anomal_util.anomalib.utils.metrics.aupro', 'AI_Module.anomal_util.anomalib.utils.metrics.auroc', 'AI_Module.anomal_util.anomalib.utils.metrics.binning', 'AI_Module.anomal_util.anomalib.utils.metrics.collection', 'AI_Module.anomal_util.anomalib.utils.metrics.min_max', 'AI_Module.anomal_util.anomalib.utils.metrics.optimal_f1', 'AI_Module.anomal_util.anomalib.utils.metrics.plotting_utils', 'AI_Module.anomal_util.anomalib.utils.metrics.pro', 'AI_Module.anomal_util.anomalib.utils.sweep.config', 'AI_Module.anomal_util.anomalib.utils.sweep.helpers.callbacks', 'AI_Module.anomal_util.anomalib.utils.sweep.helpers.inference', 'AI_Module.anomal_util.tests.helpers.aupro_reference', 'AI_Module.anomal_util.tests.helpers.config', 'AI_Module.anomal_util.tests.helpers.dataset', 'AI_Module.anomal_util.tests.helpers.detection', 'AI_Module.anomal_util.tests.helpers.dummy', 'AI_Module.anomal_util.tests.helpers.inference', 'AI_Module.anomal_util.tests.helpers.metrics', 'AI_Module.anomal_util.tests.helpers.model', 'AI_Module.anomal_util.tests.helpers.shapes', 'AI_Module.anomal_util.tests.nightly.models.test_model_nightly', 'AI_Module.anomal_util.tests.nightly.tools.benchmarking.test_benchmarking', 'AI_Module.anomal_util.tests.pre_merge.config.test_config', 'AI_Module.anomal_util.tests.pre_merge.datasets.test_bounding_box_utils', 'AI_Module.anomal_util.tests.pre_merge.datasets.test_collate', 'AI_Module.anomal_util.tests.pre_merge.datasets.test_datamodule', 'AI_Module.anomal_util.tests.pre_merge.datasets.test_dataset', 'AI_Module.anomal_util.tests.pre_merge.datasets.test_synthetic_data', 'AI_Module.anomal_util.tests.pre_merge.datasets.test_tiler', 'AI_Module.anomal_util.tests.pre_merge.datasets.test_video', 'AI_Module.anomal_util.tests.pre_merge.deploy.test_inferencer', 'AI_Module.anomal_util.tests.pre_merge.models.test_feature_extractor', 'AI_Module.anomal_util.tests.pre_merge.models.test_model_premerge', 'AI_Module.anomal_util.tests.pre_merge.models.components.test_blur', 'AI_Module.anomal_util.tests.pre_merge.post_processing.test_visualizer', 'AI_Module.anomal_util.tests.pre_merge.pre_processing.test_tiler', 'AI_Module.anomal_util.tests.pre_merge.pre_processing.transforms.test_transforms', 'AI_Module.anomal_util.tests.pre_merge.tools.conftest', 'AI_Module.anomal_util.tests.pre_merge.tools.test_gradio_entrypoint', 'AI_Module.anomal_util.tests.pre_merge.tools.test_lightning_entrypoint', 'AI_Module.anomal_util.tests.pre_merge.tools.test_openvino_entrypoint', 'AI_Module.anomal_util.tests.pre_merge.tools.test_test_entrypoint', 'AI_Module.anomal_util.tests.pre_merge.tools.test_torch_entrypoint', 'AI_Module.anomal_util.tests.pre_merge.tools.test_train_entrypoint', 'AI_Module.anomal_util.tests.pre_merge.utils.test_config', 'AI_Module.anomal_util.tests.pre_merge.utils.callbacks.export_callback.dummy_lightning_model', 'AI_Module.anomal_util.tests.pre_merge.utils.callbacks.export_callback.test_export', 'AI_Module.anomal_util.tests.pre_merge.utils.callbacks.metrics_configuration_callback.test_metrics_configuration_callback', 'AI_Module.anomal_util.tests.pre_merge.utils.callbacks.normalization_callback.test_normalization_callback', 'AI_Module.anomal_util.tests.pre_merge.utils.callbacks.visualizer_callback.dummy_lightning_model', 'AI_Module.anomal_util.tests.pre_merge.utils.callbacks.visualizer_callback.test_visualizer', 'AI_Module.anomal_util.tests.pre_merge.utils.loggers.test_get_logger', 'AI_Module.anomal_util.tests.pre_merge.utils.metrics.test_adaptive_threshold', 'AI_Module.anomal_util.tests.pre_merge.utils.metrics.test_aupro', 'AI_Module.anomal_util.tests.pre_merge.utils.metrics.test_binning', 'AI_Module.anomal_util.tests.pre_merge.utils.metrics.test_pro', 'AI_Module.anomal_util.tests.pre_merge.utils.sweep.conftest', 'AI_Module.anomal_util.tests.pre_merge.utils.sweep.test_config', 'AI_Module.anomal_util.tests.pre_merge.utils.sweep.test_throughput', 'AI_Module.anomal_util.tools.test', 'AI_Module.anomal_util.tools.train', 'AI_Module.anomal_util.tools.benchmarking.benchmark', 'AI_Module.anomal_util.tools.benchmarking.utils.metrics', 'AI_Module.anomal_util.tools.hpo.sweep', 'AI_Module.anomal_util.tools.inference.gradio_inference', 'AI_Module.anomal_util.tools.inference.lightning_inference', 'AI_Module.anomal_util.tools.inference.openvino_inference', 'AI_Module.anomal_util.tools.inference.torch_inference', 'AI_Module.RT_DETR_util.RT_DETR_Tool', 'AI_Module.RT_DETR_util.src.core.config', 'AI_Module.RT_DETR_util.src.core.yaml_config', 'AI_Module.RT_DETR_util.src.core.yaml_utils', 'AI_Module.RT_DETR_util.src.data.dataloader', 'AI_Module.RT_DETR_util.src.data.functional', 'AI_Module.RT_DETR_util.src.data.transforms', 'AI_Module.RT_DETR_util.src.data.coco.coco_dataset', 'AI_Module.RT_DETR_util.src.data.coco.coco_eval', 'AI_Module.RT_DETR_util.src.data.coco.coco_utils', 'AI_Module.RT_DETR_util.src.misc.dist', 'AI_Module.RT_DETR_util.src.misc.logger', 'AI_Module.RT_DETR_util.src.misc.visualizer', 'AI_Module.RT_DETR_util.src.nn.arch.classification', 'AI_Module.RT_DETR_util.src.nn.backbone.common', 'AI_Module.RT_DETR_util.src.nn.backbone.presnet', 'AI_Module.RT_DETR_util.src.nn.backbone.test_resnet', 'AI_Module.RT_DETR_util.src.nn.backbone.utils', 'AI_Module.RT_DETR_util.src.nn.criterion.utils', 'AI_Module.RT_DETR_util.src.optim.amp', 'AI_Module.RT_DETR_util.src.optim.ema', 'AI_Module.RT_DETR_util.src.optim.optim', 'AI_Module.RT_DETR_util.src.solver.det_engine', 'AI_Module.RT_DETR_util.src.solver.det_solver', 'AI_Module.RT_DETR_util.src.solver.solver', 'AI_Module.RT_DETR_util.src.zoo.rtdetr.box_ops', 'AI_Module.RT_DETR_util.src.zoo.rtdetr.denoising', 'AI_Module.RT_DETR_util.src.zoo.rtdetr.hybrid_encoder', 'AI_Module.RT_DETR_util.src.zoo.rtdetr.matcher', 'AI_Module.RT_DETR_util.src.zoo.rtdetr.rtdetr', 'AI_Module.RT_DETR_util.src.zoo.rtdetr.rtdetr_criterion', 'AI_Module.RT_DETR_util.src.zoo.rtdetr.rtdetr_decoder', 'AI_Module.RT_DETR_util.src.zoo.rtdetr.rtdetr_postprocessor', 'AI_Module.RT_DETR_util.src.zoo.rtdetr.utils', 'AI_Module.RT_DETR_util.tools.export_onnx', 'AI_Module.RT_DETR_util.tools.train'],

    classifiers=[
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
    ],
    package_dir={},
)
